/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     check.c                                                        */
/*                                                                          */
/* description:  checks on the initial grid and utilities                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

#if DIM == 1
static int mod_2[12] = {0,1,0,1,0,1,0,1};
static int *mod = mod_2 + 2;
#endif

#if DIM == 2
static int mod_3[12] = {0,1,2,0,1,2,0,1,2};
static int *mod = mod_3 + 3;
#endif

#if DIM == 3
static int mod_4[12] = {0,1,2,3,0,1,2,3,0,1,2,3};
static int *mod = mod_4 + 4;
#endif

static int  error_detected;

static void check_fct(const EL_INFO *el_info)
{
  FUNCNAME("check_fct");
  int       i, j, k, opp_v;
  EL        *el, *neigh;

#if !NEIGH_IN_EL
  TEST_FLAG(FILL_NEIGH, el_info);
#endif

  el = el_info->el;
  for (i = 0; i < N_NEIGH; i++)
  {
    if ((neigh = NEIGH(el, el_info)[i]))
    {
#if DIM == 1
      if (el_info->bound[1-i])
#else
      if (el_info->boundary[i])
#endif
      {
	if (!error_detected)
	  MSG("error detected!!!\n");
	error_detected++;
	MSG("interior boundary[%d] non nil on element = %d\n", i, INDEX(el));
      }

      opp_v = OPP_VERTEX(el,el_info)[i];
      if(opp_v < 0  ||  opp_v >= N_NEIGH)
      {
	if (!error_detected)
	  MSG("error detected!!!\n");
	error_detected++;
	MSG("opp_v = %d\n", opp_v);
      }
      for (j = 1; j < N_VERTICES; j++)
      {
	for (k = 1; k < N_VERTICES; k++)
	  if (el->dof[mod[i+j]] == neigh->dof[mod[opp_v+k]]) break;

	if (k >= N_VERTICES)
	{
	  if (!error_detected)
	    MSG("error detected!!!\n");
	  error_detected++;
	  MSG("dof %d of el %d at face %d isn't dof of neigh %d at face %d\n",
	      el->dof[mod[i+j]][0], INDEX(el), i, INDEX(neigh), opp_v);
	}	   
      }
    }
    else
    {
#if DIM == 1
      if (!el_info->bound[1-i])
#else
      if (!el_info->boundary[i])
#endif
      {
	if (!error_detected)
	  MSG("error detected!!!\n");
	error_detected++;
	MSG("boundary[%d] on domains boundary is nil on element = %d\n",
	    i, INDEX(el));
      }
    }

#if NEIGH_IN_EL==2
    if (el->child[0] == nil) 
    {
      for (i = 0; i < N_NEIGH; i++)
      {
	if (el_info->neigh[i] != el->neigh[i])
	{
	  if (!error_detected)
	    MSG("error detected!!!\n"); 
	  error_detected++;
	  MSG("el_info->neigh[%d] != el->neigh[%d] on el %d\n", i, i,
	      INDEX(el));
	}
	if (el->neigh[i])
	  if (el_info->opp_vertex[i] != el->opp_vertex[i])
	  {
	    if (!error_detected)
	      MSG("error detected!!!\n"); 
	    error_detected++;
	    MSG("el_info->opp_vertex[%d] != el->opp_vertex[%d] on el %d\n",
		i, i,INDEX(el));
	  }
      }
    }
#endif
  }

  return;
}

/*--------------------------------------------------------------------------*/
/*  check dofs from mesh->dof_admin[iadmin]                                 */
/*--------------------------------------------------------------------------*/

static int       iadmin = 0;
static int       dof_size = 0;
static int       *dof_used = nil;

static void check_dof_fct(const EL_INFO *el_info)
{
  FUNCNAME("check_dof_fct");
  MESH      *mesh  = el_info->mesh;
  DOF_ADMIN *admin = mesh->dof_admin[iadmin];
  EL        *el = el_info->el;
  DOF       *dof;
  int       i, j, jdof, ndof, i0, j0;
#if DIM > 1
  int       ov;
  EL        *neigh;
#endif

#if DIM == 3
  int       in, k, found;
#endif

  if (!dof_used) return;

  if ((ndof = admin->n_dof[VERTEX]))
  {
    j0 = admin->n0_dof[VERTEX];
    TEST_EXIT(j0 + ndof <= mesh->n_dof[VERTEX])
      ("admin->n0_dof[VERTEX] %d + n_dof %d > mesh->n_dof %d\n",
       j0, ndof, mesh->n_dof[VERTEX]);
    i0 = mesh->node[VERTEX];
    for (i = 0; i < N_VERTICES; i++)
    {
      if ((dof = el->dof[i0+i]) == nil)
	ERROR("no vertex dof %d on element %d\n", i, INDEX(el_info->el));
      else
	for (j = 0; j < ndof; j++)
	{
	  jdof = dof[j0 + j];
	  TEST(jdof >= 0 && jdof < dof_size)
	    ("vertex dof=%d invalid? size=%d\n",jdof, dof_size);
	  dof_used[jdof]++;
	}
    }
    /* neighbour vertex dofs have been checked in check_fct() */
  }

#if DIM > 1
  if ((ndof = admin->n_dof[EDGE]))
  {
    j0 = admin->n0_dof[EDGE];
    TEST_EXIT(j0 + ndof <= mesh->n_dof[EDGE])
      ("admin->n0_dof[EDGE] %d + n_dof %d > mesh->n_dof %d\n",
       j0, ndof, mesh->n_dof[EDGE]);
    i0 = mesh->node[EDGE];
    for (i = 0; i < N_EDGES; i++)
    {
      if ((dof = el->dof[i0 + i]) == nil)
	ERROR("no edge dof %d on element %d\n", i, INDEX(el_info->el));
      else
	for (j = 0; j < ndof; j++)
	{
	  jdof = dof[j0 + j];
	  TEST(jdof >= 0 && jdof < dof_size)
	    ("edge dof=%d invalid? size=%d\n",jdof, dof_size);
	  dof_used[jdof]++;
	}
      if (el->child[0] == nil)
      {
#if DIM==2
	if ((neigh = NEIGH(el, el_info)[i]))
	{
	  ov = OPP_VERTEX(el, el_info)[i];
	  TEST(neigh->dof[i0 + ov] == dof)
	    ("el %d edge %d dof %8X: wrong dof %8X in neighbour %d edge %d\n",
	     INDEX(el), i, dof, neigh->dof[i0 + ov], INDEX(neigh), ov);
	}
#else
	for (in = 0; in < N_NEIGH; in++)
	{
	  if ((in != vertex_of_edge[i][0]) && (in != vertex_of_edge[i][1])
	      && (neigh = NEIGH(el, el_info)[in]))
	  {
	    found = 0;
	    for (k = 0; k < N_EDGES; k++)
	      if (neigh->dof[i0 + k] == dof) found++;
	    TEST(found==1)("el %d edge %d dof found=%d in neighbour %d\n",
			   INDEX(el), i, found, INDEX(neigh));
	  }
	}
#endif
      }
    }
  }
#endif

#if DIM==3
  if ((ndof = admin->n_dof[FACE]))
  {
    j0 = admin->n0_dof[FACE];
    TEST_EXIT(j0 + ndof <= mesh->n_dof[FACE])
      ("admin->n0_dof[FACE] %d + n_dof %d > mesh->n_dof %d\n",
       j0, ndof, mesh->n_dof[FACE]);
    i0 = mesh->node[FACE];
    for (i = 0; i < N_FACES; i++)
    {
      TEST(dof = el->dof[i0 + i])("no face dof %d ???\n", i);
      for (j = 0; j < ndof; j++)
      {
	jdof = dof[j0 + j];
	TEST(jdof >= 0 && jdof < dof_size)
	  ("face dof=%d invalid? size=%d\n",jdof, dof_size);
	dof_used[jdof]++;
      }

      if (el->child[0] == nil)
      {
	if ((neigh = NEIGH(el, el_info)[i]))
	{
	  ov = OPP_VERTEX(el, el_info)[i];
	  TEST(neigh->dof[i0 + ov] == dof)
	    ("el %d face %d dof %8X: wrong dof %8X in neighbour %d face %d\n",
	     INDEX(el), i, dof, neigh->dof[i0 + ov], INDEX(neigh), ov);
	}
      }
    }
  }
#endif

  if ((ndof = admin->n_dof[CENTER]))
  {
    i0 = mesh->node[CENTER];
    TEST(dof = el->dof[i0])("no center dof???\n");
    j0 = admin->n0_dof[CENTER];
    TEST_EXIT(j0 + ndof <= mesh->n_dof[CENTER])
      ("admin->n0_dof[CENTER] %d + n_dof %d > mesh->n_dof %d\n",
       j0, ndof, mesh->n_dof[CENTER]);
    for (j = 0; j < ndof; j++)
    {
      jdof = dof[j0 + j];
      TEST(jdof >= 0 && jdof < dof_size)
	("center dof=%d invalid? size=%d\n",jdof, dof_size);
      dof_used[jdof]++;
    }
  }

  return;
}

/*--------------------------------------------------------------------------*/

static void node_fct(const EL_INFO *el_info)
{
  FUNCNAME("node_fct");
  int     i;
  EL      *el = el_info->el;

  MSG("el %4d: ", INDEX(el));
  for (i = 0; i < N_VERTICES; i++)
    print_msg("%4d%s", el->dof[i][0], i < N_VERTICES-1 ? ", " : "\n");

  return;
}

/*--------------------------------------------------------------------------*/

void check_mesh(MESH *mesh)
{
  FUNCNAME("check_mesh");
  int       i, nused, nfree;
  DOF_ADMIN *admin;
  FLAGS     fill_flag;
  int       info = -1;

  if (info < 0)
  {
    info = 2;
    GET_PARAMETER(0, "refine/coarsen info", "%d", &info);
  }

  if (info >= 2)
    MSG("checking mesh\n");

  error_detected = 0;


#if NEIGH_IN_EL
  fill_flag = CALL_LEAF_EL | FILL_BOUND;
#else
  fill_flag = CALL_EVERY_EL_INORDER | FILL_NEIGH | FILL_BOUND;
#endif
  mesh_traverse(mesh, -1, fill_flag, check_fct);

  if (mesh->preserve_coarse_dofs)
    fill_flag = CALL_EVERY_EL_INORDER;
  else 
    fill_flag = CALL_LEAF_EL;
  
  if (!NEIGH_IN_EL)
    fill_flag |= FILL_NEIGH;

  for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
  {
    TEST_EXIT(admin = mesh->dof_admin[iadmin])
      ("no dof_admin[%d]\n", iadmin);
    if (admin->size > 0)
    {
      if (dof_size < admin->size)
      {
	dof_used = MEM_REALLOC(dof_used, dof_size, admin->size + 1000, int);
	dof_size = admin->size + 1000;
      }
      for (i = 0; i < dof_size; i++) dof_used[i] = 0;

      nused = nfree = 0;

      mesh_traverse(mesh, -1, fill_flag, check_dof_fct);

      FOR_ALL_DOFS(admin, 
		   nused++;
		   if (!dof_used[dof]) {
		     error_detected++;
		     MSG("dof[%d] not used??\n",dof);
		   });

      FOR_ALL_FREE_DOFS(admin,
			nfree++;
			if (dof_used[dof]) {
			  error_detected++;
			  MSG("dof[%d] used??\n",dof);
			});

      TEST(nused + nfree == admin->size)
	("nused = %d, nfree = %d, admin->size = %d ????\n",
	 nused, nfree, admin->size);
      TEST(nused == admin->used_count)
	("nused = %d, admin->used_count = %d ?????\n",
	 nused, admin->used_count);
    }
  }

  MEM_FREE(dof_used, dof_size, int);
  dof_used = nil;
  dof_size = 0;

  if (!error_detected)
  {
    if (info >= 2)
      MSG("checking done; no error detected\n");
  }
  else
  {
    MSG("checking done; %d error%s detected\n", error_detected,
	error_detected == 1 ? "" : "s");
    mesh_traverse(mesh, -1, CALL_EVERY_EL_INORDER, node_fct);
    WAIT_REALLY;
  }

  return;
}

static const DOF  *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
static int        n_bas_fcts;
static const DOF_ADMIN  *admin;

/*  scalar problems                                                        */
static REAL       *inter_vec, (*inter_fct)(const REAL_D);
static const REAL *(*interpol_el)(const EL_INFO *, int, const int *, 
				    REAL (*)(const REAL_D), REAL *);

/*  vector valued problems                                                 */
static REAL_D     *inter_vec_d;
static const REAL *(*inter_fct_d)(const REAL_D, REAL_D);
static const REAL_D *(*interpol_el_d)(const EL_INFO *, int, const int *b_no,
				      const REAL *(*)(const REAL_D, REAL_D), 
				      REAL_D *);

static void interpol_fct(const EL_INFO *el_info)
{
  int         i;
  const DOF   *dof = get_dof_indices(el_info->el, admin, nil);
  const REAL  *inter_val = interpol_el(el_info, 0, nil, inter_fct, nil);

  for (i = 0; i < n_bas_fcts; i++)
    inter_vec[dof[i]] = inter_val[i];

  return;
}

void interpol(REAL (*fct)(const REAL_D), DOF_REAL_VEC *vec)
{
  FUNCNAME("interpol");
  const FE_SPACE   *fe_space;

  GET_DOF_VEC(inter_vec, vec);

  if (!(fe_space = vec->fe_space))
  {
    MSG("no dof admin in vec %s, skipping interpolation\n", NAME(vec));
    return;
  }

  if (!(admin = fe_space->admin))
  {
    MSG("no dof admin in fe_space %s, skipping interpolation\n",
	NAME(fe_space));
    return;
  }
  
  if (!fe_space->bas_fcts)
  {
    MSG("no basis functions in admin of vec %s, skipping interpolation\n", 
	NAME(vec));
    return;
  }
  n_bas_fcts = fe_space->bas_fcts->n_bas_fcts;

  if (!(inter_fct = fct))
  {
    MSG("function that should be interpolated only pointer to nil, ");
    print_msg("skipping interpolation\n");
    return;
  }

  if (!(interpol_el = fe_space->bas_fcts->interpol))
  {
    MSG("no function for interpolation on an element available\n");
    MSG("in basis functions of vec %s, skipping interpolation\n", NAME(vec));
    return;
  }
  if (!(get_dof_indices = fe_space->bas_fcts->get_dof_indices))
  {
    MSG("no function for getting dof's on an element available\n");
    MSG("in basis functions of vec %s, skipping interpolation\n", NAME(vec));
    return;
  }

  mesh_traverse(fe_space->mesh, -1, CALL_LEAF_EL|FILL_COORDS, interpol_fct);

  return;
}

static void interpol_fct_d(const EL_INFO *el_info)
{
  int          i, k;
  const DOF    *dof = get_dof_indices(el_info->el, admin, nil);
  const REAL_D *inter_val = interpol_el_d(el_info, 0, nil, inter_fct_d, nil);

  for (i = 0; i < n_bas_fcts; i++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      inter_vec_d[dof[i]][k] = inter_val[i][k];

  return;
}

void interpol_d(const REAL *(*fct)(const REAL_D, REAL_D), DOF_REAL_D_VEC *vec)
{
  FUNCNAME("interpol_d");
  const FE_SPACE   *fe_space;

  GET_DOF_VEC(inter_vec_d, vec);

  if (!(fe_space = vec->fe_space))
  {
    MSG("no dof admin in vec %s, skipping interpolation\n", NAME(vec));
    return;
  }

  if (!(admin = fe_space->admin))
  {
    MSG("no dof admin in fe_space %s, skipping interpolation\n",
	NAME(fe_space));
    return;
  }
  
  if (!fe_space->bas_fcts)
  {
    MSG("no basis functions in admin of vec %s, skipping interpolation\n", 
	NAME(vec));
    return;
  }
  n_bas_fcts = fe_space->bas_fcts->n_bas_fcts;

  if (!(inter_fct_d = fct))
  {
    MSG("function that should be interpolated only pointer to nil, ");
    print_msg("skipping interpolation\n");
    return;
  }

  if (!(interpol_el_d = fe_space->bas_fcts->interpol_d))
  {
    MSG("no function for interpolation on an element available\n");
    MSG("in basis functions of vec %s, skipping interpolation\n", NAME(vec));
    return;
  }
  if (!(get_dof_indices = fe_space->bas_fcts->get_dof_indices))
  {
    MSG("no function for getting dof's on an element available\n");
    MSG("in basis functions of vec %s, skipping interpolation\n", NAME(vec));
    return;
  }

  mesh_traverse(fe_space->mesh, -1, CALL_LEAF_EL|FILL_COORDS, interpol_fct_d);

  return;
}
